###################################
# PROLOG SECTION
# blackjack.py
# A program that plays a simplified version of blackjack (21).
#
# This program plays a simplified version of the game
# of blackjack, or 21. In this version of the game
# the rules are these:
#
# * The cards each have a count. Cards 2-10 have a
#   count of the face value. Jack-king have a count
#   of 10 also. An ace can be worth either 1 or 11.
#
# * The count of a hand is the total count of all cards
#   in the hand. If a hand has a value over 21 and it
#   contains an ace, the ace can be counted as a 1 instead.
#
# * The winner of each hand is the player whose total
#   count is closest to 21. If neither player has 21,
#   the player with the higher total count wins.
#   If both players have the same count, the hand is a tie.
#   Any player whose hand has a value over 21, is 'busted'
#   and automatically loses.
#
# * At the start of the game, the player is dealt
#   two cards, and the dealer is dealt two cards.
#   The first card dealt to the dealer is displayed
#   to the player.
#
# * The player goes first and can elect to get more cards.
#   If the value of the player's hand exceeds 21,
#   she loses immediately.
#
# * Once the player stops taking any more cards, it is the
#   dealer's turn. The dealer must take a new card as long
#   as the value of his hand is less than 17. Once it reaches
#   or exceeds 17, he stops. If the value of the dealer's hand
#   exceeds 21, he lose the hand.
#
# Tested with Python 3.7.9 on Windows 10
# Fri July 16, 2021
#
# Copyright ©. All rights reserved. NAF 2021.
#
# Possible future enhancements: betting, doubling hands
# Unresolved bugs:
###################################

###################################
# ENVIRONMENT SETUP SECTION
# Imports: random
###################################

# import random number routines
import random

###################################
# FUNCTION DEFINITION SECTION
# Definitions: card_display()
#  show_cards(), shuffle_deck()
#  card_value(), hand_value()
#  deal_from_deck(), deal_card()
###################################

#
# Function to define the cards in the deck.
#
# This function is nearly identical to define_cards.py,
# which should be used as a model.
# The function assigns string values to a list of
# integer variables representing the cards (52 in all).
# The list items range from 0 to 51.
# The value strings include both card rank
# (e.g., seven or queen) and suit (e.g., clubs or hearts).
# Examples: List item 0 represents string "ace of clubs";
# list item 2 represents "three of clubs";
# list item 51 represents "king of spades".
#
def define_cards(n):
    # debug printing
    if debug_print:
        print("DEBUG: enter define_cards with",n)
    # strings for card ranks
    rank_string = ("ace","two","three","four","five","six","seven", \
                "eight","nine","ten","jack","queen","king")
    # strings for card suits
    suit_string = ("clubs","diamonds","hearts","spades")
    # initialize the list of cards to null
    cards = []
    # build up the strings for all cards in the deck
    # begin with the suits
    for suit in range(4):
        # first go through all ranks for clubs, then diamonds, etc.
        for rank in range(13):
            # assign the card string (e.g., "ace of clubs")
            # to rank_string
            card_string = rank_string[rank] + " of " + suit_string[suit]
            # append each card_string in turn
            # to the end of the list of cards
            cards.append(card_string)

    # return the definition of the appropriate card
    # to the calling function
    return cards[n]

#
# Function to return the content of a card
# as a string including suit and rank and value.
# Note: a card is an integer from 0-51.
#
# This function calls define_cards() and
# card_value() to create the string it returns
# to the caller. This allows us to reuse define-cards(),
# which we wrote in an earlier lesson.
#
def card_display(n):
    # debug printing
    if debug_print:
        print("DEBUG: enter card_display with",n)

    # get the card definition from define_cards()
    card_definition = define_cards(n)

    # get the value of the card as an integer
    value = card_value(n)
    
    # return the rank, suit and value as a single string
    return card_definition+" ("+str(value)+")"

#
# Function to return the value of a card
# an ace is either 1 or 11, and 10-king are all 10.
#
# We use the rank of the card (ace-king) to look it's
# point value up in the card_values tuple.
#
def card_value(n):
    # debug printing
    if debug_print:
        print("DEBUG: enter card_value with",n)
        
    # tuple of card values ace-king
    card_values=(11,2,3,4,5,6,7,8,9,10,10,10,10)
    # get the rank of the card
    card_rank = n % 13
    # get the point value of that rank
    card_value = card_values[card_rank]
    # return the point value
    return card_value

#
# Function to return the value of a hand
#
# We start by calling card_value() for each
# card in the hand and adding them all up.
# We also count how many aces are in the hand,
# since they can be valued as either 1 or 11.
#
# If the hand has a value over 21, we recalculate
# aces as 1 instead of 11 until we (hopefully)
# get to hand value less than 21.
#
def hand_value(hand):
    # debug printing
    if debug_print:
        print("DEBUG: enter hand_value with",hand)

    # initialize the total hand value and the number of aces
    value=0
    aces=0
    # add up the value of each of the cards
    for card in hand:
        # call card_value() to get the value of this card
        cvalue=card_value(card)
        # count the number of aces (value 11) that are in the hand
        if cvalue==11:
            aces=aces+1
        # add to the total hand value
        value=value+cvalue

    # if the total hand value is greater than 21, try counting
    # any aces as 1 to reduce the value to be below 21.
    if value>21 and aces>0:
        for i in range(aces):
            if value>21:
                value=value-10
        
    return value

#
# Function to print out a list of cards
# by displaying the value and suit for each card.
#
# We do this by calling card_display() for each card
# in the hand and then displaying the result with
# three cards in each line of output.
#
def show_cards(mydeck):
    # debug printing
    if debug_print:
        print("DEBUG: enter show_cards with",mydeck)

    # initialize card counter
    j=0
    # loop through the cards in the hand
    for card in mydeck:
        # print out the card_display() string
        # for the card left justified.
        print(card_display(card).ljust(20)," ",)
        # add a new line every three cards
        if j%3 == 2:
            print()
        # add to the card counter
        j=j+1
    print()

#
# Function to create a deck.
#
# We create the deck as a list of numbers
# [0,1,2, ... , 51].
#
def create_deck(deck):
    # debug printing
    if debug_print:
        print("DEBUG: create_deck")
        
    # the deck was initialized to null in main program
    # create a deck that consists solely of 52 numbers
    # (0-51)
    # deck is sequenced like a new pack of cards
    # (i.e., rank-ordered within suit)
    # there is as yet no correspondence between the numbers
    # and the values established in define_cards
    for i in range(52):
        deck.append(i)
    # return to the calling function
    return

#
# Function to shuffle the deck.
#
# Call the random.shuffle function to do the shuffle.
#
# This function shuffles its list argument in place.
#
# After this call, the list will be in a random order
# rather than being in strict sequence. It will be
# something like [2,45,88, ... ,0]
#
def shuffle_deck(deck):
    # debug printing
    if debug_print:
        print("DEBUG: enter shuffle_deck")
    random.shuffle(deck)
    # return to the calling function
    return

#
# Function to "deal" a card by returning and removing the
# first card in the deck list.
#
# (in this case, simply turn a card over to reveal its face).
#
def deal_card(deck):
    # debug printing
    if debug_print:
        print("DEBUG: enter deal_card")
        
    # return the first card in the deck to the calling function
    # hint: to deal from the bottom of the deck,
    # change the statement to: deck.pop()
    # hint: to see the effect, put in BOTH "print deck"
    # statements in the processing section 
    return deck.pop(0)

#
# Function to deal a card and if necessary, create and shuffle
# a new deck.
#
# This function acts as a wrapper for several other functions
# we already have. In addition to just returning the first card
# in the deck, it also handles the case when we have run out of
# cards by creating a new deck, shuffling it, and then returning
# the top card. If you play blackjack long enough, the deck can
# run out of cards while you are playing the game.
#
def deal_from_deck(deck):
    # debug printing
    if debug_print:
        print("DEBUG: enter deal_from_deck")
    # if the deck is empty, create a new shuffled deck
    if len(deck)<1 :
        create_deck(deck)
        shuffle_deck(deck)

    # pick a card
    thecard = deal_card(deck)
    
    # debug printing
    if debug_print:
        print("DEBUG: deal_from_deck returns",thecard)
        
    # return the first card in the deck
    return thecard


#    
# Function to print out the rules
# of the game.
#
def print_game_rules():
    print()
    print("Welcome to the game of blackjack.")
    print("You are playing against the dealer.")
    return

###################################
# PROCESSING INITIALIZATION SECTION
###################################

# define global card deck constant
values=13
suits=4
cards=values*suits

draws=0
player_wins=0
dealer_wins=0

# set the value of the debug print flag
debug_print = False

###################################
# PROCESSING SECTION
#
# 1. First deal two cards each to the player
#    and the dealer. Display the dealer's top
#    card.
# 2. Deal cards to the player until they stop
#    drawing or they go over 21.
# 3. Deal cards to the dealer until their hand
#    value reaches 17 or they go over 21.
# 4. Compare the value of the hands to see who
#    won the hand.
# 5. Ask the player if they want to play another hand.
# 6. Display statistics on hands won.
#
# Branching code: if, if-else, if-else-elif
# Looping code: for-loop, while-loop
###################################

# print the rules of the game
print_game_rules()

# initialized to an empty deck
deck=[]

# main loop to play a series of hands
keep_playing = True

while keep_playing:
    # start a new hand
    print()
    print("*** New hand ***")
    print()
    
    # initialize the two hands
    player=[]
    dealer=[]
    # deal two cards to the player and two to the dealer
    player.append(deal_from_deck(deck))
    player.append(deal_from_deck(deck))
    dealer.append(deal_from_deck(deck))
    dealer.append(deal_from_deck(deck))
    
    # debug printing
    if debug_print:
        print("DEBUG: dealer hand")
        show_cards(dealer)
        print("DEBUG: player hand")
        show_cards(player)
        
    # display the dealer's up card
    print()
    print("The dealer is showing:",card_display(dealer[0]))

    # deal cards to the player until her hand value is
    # over 21 or she stop drawing cards
    keep_dealing = True
    while keep_dealing:
        # display the player hand
        print()
        # remember player hand value
        player_has = hand_value(player)
        print("Your hand is:",)
        print(player_has)
        show_cards(player)
        # does the player already have 21?
        if player_has == 21:
            print("You have 21!")
        # if the player is not busted, ask them
        # if they want to draw a card
        if player_has < 21:
            # see if the player wants another card
            ans = input("Do you want another card? (y/n) ")
            ans = ans.lower()
            if ans == 'y':
                # draw another card for the player
                player.append(deal_from_deck(deck))
            else:
                # the player does not want to draw another card
                keep_dealing = False
        else:
            # stop because the player is busted or has 21
            keep_dealing = False

    # debug printing
    if debug_print:
        print("DEBUG: player has")
        show_cards(player)
        
    # if the player is not busted, it is the dealer's turn
    if player_has<22:
        # now it is the dealer's turn to draw cards
        keep_drawing = True
        # we use a while loop here, because we don't know
        # how many cards the dealer will draw
        while keep_drawing:
             # display what the dealer has
             print()
             dealer_has = hand_value(dealer)
             print("The dealer hand is:",dealer_has)
             show_cards(dealer)
             # the dealer keeps drawing until he reaches at least 17
             if dealer_has<17:
                 print("Dealer draws another card.")
                 dealer.append(deal_from_deck(deck))
             else:
                 keep_drawing = False
                

    # determine who wins this hand
    print()
    # get the two hand values
    player_has=hand_value(player)
    dealer_has=hand_value(dealer)
    print("You have",player_has,"and the dealer has",dealer_has)

    # now score the hand by who has the highest value not
    # over 21.
    
    # did the player go bust? if she did, the dealer
    # didn't draw anything.
    if player_has > 21:
        print("You are busted, the dealer wins.")
        dealer_wins=dealer_wins+1
    # did the dealer go bust?
    elif dealer_has > 21:
        print("The dealer is busted, you win.")
        player_wins=player_wins+1
    # was the hand a tie?
    elif player_has == dealer_has:
        print("This hand is a draw")
        draws=draws+1
    # did the dealer win?
    elif player_has < dealer_has:
        print("The dealer wins.")
        dealer_wins=dealer_wins+1
    # only other possibility, the player won.
    else:
        print("You win.")
        player_wins=player_wins+1

    # does player want to play another hand?
    print()
    ans = input("Do you want to play another hand? (y/n) ")
    ans = ans.lower()
    if ans != 'y':
       keep_playing = False
    

###################################   
# CLEANUP, TERMINATION, AND EXIT SECTION
###################################

print()
print("End blackjack.")
print()

# display the score
print("You won",player_wins,"hands")
print("The dealer won",dealer_wins,"hands")
print(draws,"hands were a draw")
input()